Unit U_StarsOnAGrid;
 {Copyright   2002-2017, Gary Darby,  www.DelphiForFun.org
 This program may be used or modified for any non-commercial purpose
 so long as this original notice remains in place.
 All other rights are reserved
 }

 {This program uses the same efficient algorithm as the "8 Queens Problem"
 discovered by Niklaus Wirth which finds solutions very rapidly. (Niklaus Wirth,
 Algorithms and Data Structures,  1978, 1986).    In this version, we need to
 place N "Stars" in an NxN grid which has been divided into N sections in such a
 way that there is exactly 1 star in each section.  No two stars can reside in
 the same row or column, nor is any star diagonally adjacent to another star.
 }

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Grids, ComCtrls, Spin, Inifiles, ExtCtrls, dffutils, types;

type
  tsolution=array[1..10] of byte;
  tboard= array[1..10] of TSolution;

  TMode=(AutoSolving, Usersolving, UserBuilding, Unknown);

  Tflist = class(TForm)
    Panel1: TPanel;
    Label2: TLabel;
    CaseLbl: TLabel;
    CurrentColor: TShape;
    TitleLbl: TLabel;
    Label1: TLabel;
    StarImage: TImage;
    Label3: TLabel;
    WirthBtn: TButton;
    ListBox1: TListBox;
    StringGrid2: TStringGrid;
    Memo1: TMemo;
    StarCountSpin: TSpinEdit;
    SaveBtn: TButton;
    ColorGrid: TDrawGrid;
    Memo2: TMemo;
    CaseInBox: TComboBox;
    Memo3: TMemo;
    Memo4: TMemo;
    MsgBar: TStaticText;
    ResettBtn: TButton;
    StaticText2: TStaticText;
    procedure WirthBtnClick(Sender: TObject);  {Find all solutions}
    procedure ListBox1Click(Sender: TObject);
    procedure CountChange(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ResetBtnClick(Sender: TObject);
    procedure SaveBtnClick(Sender: TObject);
    procedure StringGrid2DrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure StringGrid2Exit(Sender: TObject);
    procedure ColorGridDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure ColorGridClick(Sender: TObject);
    procedure StringGrid2Click(Sender: TObject);
    procedure StringGrid2MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure CaseInBoxClick(Sender: TObject);
  public
    Grow:array[1..10] of boolean;
    ColPlusRow:array[2..20] of boolean;
    ColMinusRow:array[-10..+10] of boolean;
    solution:Tsolution;
    solutions:array of Tsolution;
    dir:string;
    StarCount:integer;
    Casename, Title:string;
    modified:boolean;  {board letters changed}
    InitialLoad:boolean;  {Flag to prevent asking about saving previous case on first load}
    mode:TMode;  {May help in preserving user star entries after accidental left click}
    procedure LoadCaseList; {call initallially and after each case save}
    procedure Search (column: integer);
    Procedure Initialize;
    procedure ShowSolution;
    Procedure SaveCase;
    Procedure LoadCase(name:string);
    function checkmodified:boolean;
    function ValidateBoardLetters:boolean;
    procedure resetSolutionDisplay;
    Procedure resetsolution;
    function starsPlaced:integer;
  end;



var
  flist: Tflist;

implementation

uses math;

{$R *.DFM}

CONST
  DefaultCase: array [0..5] of STRING = ('AAAAAA',
                                         'ABBBAB',
                                         'ABCBBB',
                                         'DDCBBB',
                                         'DECFFF',
                                         'DEEEFF');
   Colors: array ['A'..'J'] of tcolor =
           (clBlue,clYellow,clRed,clGreen,clLime,
            clMaroon,clPurple,clAqua,clFuchsia,clNavy) ;

{*++++******** RotatGrid90Right **********8}
Procedure RotateGrid90Right(grid:TstringGrid);
var
  tempgrid:array of array of string;
  maxc,maxr:integer;
  c,r:integer;
begin
  with grid do
  begin
    setlength(tempgrid, colCount, RowCount);

    maxc:=colCount-1;
    maxr:=rowCount-1;

    for c:=0 to maxc do for r:=0 to maxr do tempgrid[maxr-r,c]:= cells[c,r];
    rowcount:=maxc+1;
    colcount:=maxr+1;
    c:=defaultcolwidth;
    r:=defaultrowheight;
    defaultcolwidth:=r;
    defaultrowheight:=c;

    for c:=0 to maxc do for r:=0 to maxr do cells[c,r]:=tempgrid[c,r];
  end;
end;

{************* Search ************}
 procedure Tflist.Search (column: integer);
 {Apply Wirth solutio n search algorithm }

         {------------SetQueen -------------}
        procedure SetQueen (row, column: integer);
          begin
            gRow[row] := false;
            ColPlusRow[column + row] := false;
            ColMinusRow[column - row] := false;
          end;

        {------------ RemoveQueen ----------}
        procedure RemoveQueen (row, column: integer);
          begin
            gRow[row] := true;
            ColPlusRow[column + row] := true;
            ColMinusRow[column - row] := true;
          end;

        {-------------- AdjacentDiagonalOK -----------}
        function adjacentdiagonalOK(column,row:integer):boolean;
        begin
          result:=true;
          if (column>1) and  (abs(row-solution[column-1])=1) then result:=false;
        end;

        {------------ OneStarPerSection -----------}
        function OneStarPerSection:boolean;
        var
          i,col, row, index:integer;
          StarInSection:array of boolean;
        begin
          result:=true;
          setlength(StarInSection, starcount+1);
          for i:=-1 to starcount do StarInSection[i]:=false;
          for i:=1 to Starcount do
          begin
            col:=i-1;
            row:=solution[i]-1;
            //index:=ord(gridcopy[col,row])-ord('A')+1;
            index:=ord(stringgrid2.cells[col,row][1])-ord('A')+1;
            if not StarInSection[index] then StarInSection[index]:=true
            else
            begin
              result:=false;
              break;
            end;
          end;
        end;
        

  var
    row: integer;
    rowFlag,plusFlag,minusFlag,adjflag: Boolean;
  begin   {Search}
    for row := 1 to starCount do
    begin
      rowFlag := gRow[row];  {another queen in this row?}
      adjflag:=adjacentdiagonalOK(column,row);

      if (rowFlag and adjflag) then
      begin  {if not then add piece in the is row}
        Solution[column] := row;
        SetQueen(row, column);
        if column < Starcount then Search(column + 1)   {check next column}
        else if oneStarPerSection then ShowSolution;
        RemoveQueen(row, column);
      end;
    end;
  end;    {Search}



{************ Initialize *********}
Procedure Tflist.Initialize;
var i:integer;
begin
  for i:=low(gRow) to high(Grow) do grow[i]:=true;
  for i:=low(ColPlUsRow) to high(ColPlusRow) do ColPlusRow[i]:=true;
  for i:=low(ColMinusRow) to high(ColMinusRow) do ColMinusRow[i]:=true;

  (* Uncomment these lines to check solutions for EightQueensPlus
  {exclude main diagonals from solutions}
  ColPlusRow[9]:=false; ColMinusRow[0]:=false;
  *)

  setlength(solutions,100);
  listbox1.clear;
end;

{********* Showsolution *********}
procedure Tflist.showsolution;
var  i:integer; 
     s:string;
begin
  s:='';
  for i:=1 to starcount do s:=s+inttostr(solution[i] mod 10);
  listbox1.items.add(s);
  if listbox1.items.count>length(solutions) then setlength(solutions, length(solutions)+100);
  solutions[listbox1.items.count-1]:=solution; {save solutions}
end;


{*************** ValidateBoardLetters ************8}
function Tflist.ValidateBoardLetters:boolean;

var
  i,j:integer;
  ch:char;
  alphacounts: array['A'..'Z'] of integer;
  uniquecounts:integer;
begin
  result:=true;
  if not modified then exit;
  with stringgrid2 do
  begin

    for ch:='A' to 'Z' do alphacounts[ch]:=0;
    uniquecounts:=0;
    for i:=0 to colcount-1 do
    begin
      for j:=0 to rowcount-1 do
      begin

        //if length(cells[i,j])=0 then cells[i,j]:=' ';
        ch:=cells[i,j][1];
        (*
        if not (ch in ['A'..'Z']) then
        begin
          showmessage(format('Error in cell in row %d,column %d, board must be filled with letters',[j+1,i+1]));
          result:=false;
          exit;
        end;
        *)
        if alphacounts[ch] = 0 then inc(uniquecounts);
        inc(alphacounts[ch]);
      end;
    end;
    if uniquecounts<>starcount then
    begin
      showmessage(format('You have used %d different colors, %d are required for this %d x %d board',[uniquecounts, starcount, starcount, starcount]));
      result:=false;
    end;
    (*
    for i:=0 to colcount-1 do   for j:=0 to rowcount-1 do
    begin
      gridcopy[i,j]:=cells[i,j][1];
    end;
    *)
    modified:=true;
  end;
end;


{************** WirthBtnClick *********}
procedure Tflist.WirthBtnClick(Sender: TObject);
var s:string;
begin
  if  not validateBoardLetters  then exit;
  Initialize;
  Search(1);   {start the recursive search}
  with listbox1.items do
  if count=0 then add('No solutions found')
  else
  begin
    if count=1 then s:='' else s:='s';
    add(format(' %d solution%s found',[count,s]));
  end;
end;

{*********** FilledCount ********}
function filledcount(b:TBoard):integer;
{Count the number of squares that are occupied on the board}
var i,j:integer;
begin
  result:=0;
  for i:=1 to 10 do for j:=1 to 10 do if b[i,j]>0 then inc(result);
end;

{******** ResetSolutionDisplay ***********8}
procedure Tflist.resetSolutionDisplay;
var i,j:integer;
{Remove all stars from the display}
begin
   for i:=0 to starcount-1 do for j:=0 to starcount-1 do
  with stringgrid2 do
  if length(cells[i,j])>0 then cells[i,j]:=cells[i,j][1];
  //userstarcount:=0;
  mode:=unknown ;
end;

{********* ResetSolution ************}
Procedure TfList.resetsolution;
begin
  resetsolutionDisplay;
  listbox1.clear;
end;

{************** Listbox1Click **********}
procedure Tflist.ListBox1Click(Sender: TObject);
{Display the solution clicked by user}
var
  s,s2:string;
  i,n:integer;
begin
  if (mode=usersolving) and (starsplaced>0) and (messagedlg('You have manually placed stars on the board. Proceed anywa?',
                            mtconfirmation, [mbyes,mbno],0)=mrno) then exit;
  resetSolutionDisplay;
  sleep(1000);
  with listbox1 do
  //if itemindex>=0 then
  if not (items[itemindex][1] in ['0'..'9']) then
  begin
    showmessage('Cick a numeric line to see that solution');
    exit;
  end
  else
  begin
    s:=items[itemindex];
    for i:=1 to StarCount do
    begin
      n:=strtointdef(s[i],-1);
      if n<0 then exit;
      if n=0 then n:=10;
      with stringgrid2 do
      begin
        s2:=cells[i-1,n-1];
        if length(s2)>1 then delete(s2,length(s2),1);
        cells[i-1,n-1]:=s2+'*';
      end;
    end;
    msgbar.caption:='Program solution '+s+' displayed';
    msgbar.font.color:=clWindowtext;
    msgbar.font.style:=[];
  end;
end;



{********** CountChange **********8}
procedure Tflist.CountChange(Sender: TObject);
{Handle change in Column or Row count}
var
  mr,i,j,n:integer;
  oldcount:integer;
begin
  if modified {not initialload} then
  begin
    mr:=messagedlg('Changing size will start a new case.  Do you want to save the old case first?',
             mtConfirmation,[mbyes, mbno, mbcancel],0);
    case mr of
      mrcancel:
     begin
        StarcountSpin.value:=starcount;
        exit;
      end;
      mryes:
      begin
        savecase;
      end;
    end;
  end;
  oldcount:=starcount;
  Starcount:=StarCountSpin.value;
  stringgrid2.rowcount:=starcount;
  stringgrid2.colcount:=starcount;
  adjustgridsize(stringgrid2);

  colorgrid.RowCount:=starcount;
  colorgrid.Invalidate;
  adjustgridsize(colorgrid);
  if starcount>oldcount then
  with stringgrid2 do
  begin
    for i:=0 to colcount-1 do
    for j:=0 to rowcount-1 do
    if (i >=oldcount) or  (j>=oldcount) then
    begin
      n:=max(i,j);
      cells[i,j]:=char(ord('A')+n);
    end;
  end;
  if not initialLoad then  resetsolution;
  modified:=false;
end;


{*********** FormCreate ************}
procedure Tflist.FormCreate(Sender: TObject);
begin
  dir:=extractfilepath(application.exename);
  Casename:='Default';
  initialLoad:=true;
  ResetbtnClick(sender);
  LoadcaseList;
  mode:=unknown;
end;

{************** ResetBtnClick ***********8}
procedure Tflist.ResetBtnClick(Sender: TObject);
begin
  loadcase(casename);
end;

procedure Tflist.SaveBtnClick(Sender: TObject);
begin
  casename:=inputbox('Save Case', 'Change case name is desired:',casename);
  title:=inputbox('Title:','Change Title if desired:', title);
  savecase;
end;

{************ SaveCase **********8}
Procedure TFList.SaveCase;
var
  ini:TInifile;
  s:string;
  i,j:integer;
begin
  ini:=TInifile.create(dir+'StarsInGrid.ini');
  with ini do
  begin
    writestring(Casename, 'Title', Title);
    writeinteger(Casename, 'Starcount', Starcount);
    s:='';
    for i:=0 to Starcount-1 do
    for j:= 0  to StarCount-1 do  s:=s+stringgrid2.cells[i,j][1];{gridcopy[i,j]; }
    writestring(casename,'Cells',s);
    ini.free;
  end;
  CaseLbl.Caption:=Casename;
  TitleLbl.caption:=Title;
  msgbar.caption:=caselbl.caption+ '     '+titlelbl.caption;
  msgbar.font.color:=clWindowtext;
  msgbar.font.style:=[];
  modified:=false;
  LoadCaseList;
  CaseInBox.ItemIndex:=caseinbox.items.count-1;
end;

{********* CheckModifies **********88}
function TFList.Checkmodified:boolean;
{Call before loading a new file to ensure chance to save old case}
var
  msg:string;
  r:integer;
begin
  result:=true;
  If modified then
  begin
    msg:='Save current case first?';
    r:=messagedlg(msg,mtconfirmation,[mbyes,mbno,MBCANCEL],0);

    if r=mryes
    then
    begin
      savecase;
      Casename:='New';
      Title:='New';
      modified:=false;
    end
    else if r=mrcancel then result:=false;
  end;
end;

{************ Loadcase **********}
Procedure TFList.LoadCase(name:string);
var
  ini:TInifile;
  s:string;
  i,j:integer;
begin
  if not checkmodified then exit;
  ini:=tinifile.create(dir+'Starsingrid.ini');
  with ini do
  begin
    Casename:=name;
    Title:=readstring(Casename, 'Title', 'Unnamed');


    //if initialload or (starcountSpin.value<>starcount) then
    StarCount:=readinteger(Casename, 'Starcount', 6);
    starcountSpin.value:=starcount;
    with Stringgrid2 do
    begin
      colcount:=starcount;
      rowcount:=starcount;
      s:=readstring(casename, 'Cells', stringofchar(' ', Starcount*StarCount));
      for i:=0 to Starcount-1 do
      for j:= 0  to StarCount-1 do
      with stringgrid2 do
      begin
         //gridcopy[i,j]:=s[i*starcount+j+1];
         cells[i,j]:=s[i*starcount+j+1]; //gridcopy[i,j];
      end;
    end;
    ini.free;
  end;
  listbox1.items.clear;
  CaseLbl.caption:='Case: '+casename;
  TitleLbl.Caption:='Title: '+Title;
  msgbar.caption:=caselbl.caption+ '     ' + titlelbl.caption;
  msgbar.font.color:=clWindowtext;
  msgbar.font.style:=[];
  InitialLoad:=false;
  mode:=unknown;
end;

{******************** Stringgrid2DrawCell *****************}
procedure Tflist.StringGrid2DrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
var
  s:string;
  letter, ast:char;
begin
  With stringgrid2, canvas, rect  do
  begin
    s:=cells[acol,arow];
    letter:=s[1];
    ast:=s[length(s)];
    brush.Color:=colors[letter];
    rectangle(rect);
    if ast='*' then
    begin  {flag a solution cell}
     with rect do stretchdraw(types.rect(left, top, right, Bottom),
                  starimage.picture.bitmap);
    end;
    pen.Color:=clgray;
    pen.width:=1;
  end;
end;


{********* StringGrid2Exit ************}
procedure Tflist.StringGrid2Exit(Sender: TObject);
begin
  ValidateBoardLetters;
  msgbar.caption:=CaseLbl.caption + '    ' + Titlelbl.caption;
end;

{******** ColorGridDrawCell ***********8}
procedure Tflist.ColorGridDrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
{Fill in available colors for current grid size}
begin
  with colorgrid, canvas do
  begin
    brush.Color:=colors[char(ord('A')+arow)];
    rectangle(rect);
  end;
end;


{********** ColorGridClick ***************}
procedure Tflist.ColorGridClick(Sender: TObject);
begin
  currentcolor.Brush.Color:=colors[char(ord('A')+colorgrid.Row)];
  currentcolor.tag:=colorgrid.row;
end;


{************** StringGrid2Click **********}
procedure Tflist.StringGrid2Click(Sender: TObject);
{User clicks to change the color of a cell to the previously selected color}
var
  p:TPoint;
  acol,arow:integer;
begin
   p:=mouse.cursorpos;
  if (mode=usersolving) and (starsPlaced>0) then
  begin
    if messagedlg('You have manually added stars which will be erased. Continue?',
                 mtconfirmation, [mbyes,mbno],0) = mrno then exit;
    mode:=userbuilding;
  end;
  mode:=userbuilding;
  with Stringgrid2 do
  begin
    resetsolution; {clear any previous solution or user stars }
    if currentcolor.tag<0 then exit;  {no current color}
    {where was the click?}

    p:=screentoclient(p);
    mousetocell(p.x,p.y,acol,arow); {got it!}
    application.processmessages;
    cells[acol,arow]:=char(ord('A') + currentcolor.tag); {st the new letter (= new color)}
    modified:=true;
  end;
end;

{************ StarsPlace ************}
Function TFList.StarsPlaced:integer;
{count nbr of stars currently showing}
var
  i,j:integer;
  begin
    result:=0;
    for i := 0 to starcount-1 do for j:=0 to starcount-1 do
          if (length(stringgrid2.cells[i,j])=2) then  inc(result);
  end;

{*************** StringGrid2MouseUp *************}
procedure Tflist.StringGrid2MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
{Allow right clicks by user to add or remove stars to solve the puzzle}

  function isStar(i,j:integer):boolean;
  begin
    result:=length(stringgrid2.cells[i,j])=2;
  end;
      {------------ IsSolved ----------}
      function IsSolved:boolean;
      var
        i,j,k:integer;
        ch:char;
        rows, cols,colors:array of boolean;
        stars:integer;
        s:string;
      begin
        result:=false;
        setlength(rows,starcount);
        setlength(cols,starcount);
        setlength(colors,starcount);
        stars:=0;
        for i :=0 to starcount-1 do
        begin
          rows[i]:=false;
          cols[i]:=false;
          colors[i]:=false;
        end;
        msgbar.font.color:=clMaroon;
        msgbar.font.style:=[fsBold];
        for i := 0 to starcount-1 do for j:=0 to starcount-1 do
        begin
          s:=Stringgrid2.cells[i,j];
          ch:=s[1];
          if (length(s)=2) then
          begin
            inc(stars);
            if  cols[i] then
            begin {error}
              msgbar.Caption:=format('Multiple stars found in column %d',[i+1]);
              exit;
            end
            else cols[i]:=true;

            if rows[j] then
            begin {error}
              msgbar.Caption:=format('Multiple stars found in row %d',[j+1]);
              exit;
            end                                                              else rows[j]:=true;

            k:=ord(ch)-ord('A');
            if colors[k]then
            begin {error}
              msgbar.Caption:=format('Repeated star for color at col %d, row %d',[i+1,j+1]);
              exit;
            end
            else colors[k]:=true;

            if (j>0) and (((i>0) and  isstar(i-1,j-1))
                       or ((i<starcount-1) and  isstar(i+1,j-1)))
            then
            begin  {check up-left for an adjacent star}
              msgbar.Caption:=format('Disgonally adjacent star above col %d, row %d',[i+1,j+1]);
              exit;
            end;
          end;
        end;
        if stars=starcount then
        begin
          s:='Congratulations!  You found a solution!';
          msgbar.Caption:=s;
          msgbar.font.color:=clGreen;
          showmessage(s);
          result:=true;
        end
        else
        begin
          msgbar.font.color:=clGreen;
          msgbar.font.style:=[];
          if stars=1 then s:='' else s:='s';
          msgbar.Caption:=format('%d Star'+s+' placed OK so far.',[stars]);
        end;
      end;



var
  acol,arow:integer;
  s:string;
begin   {Stringgrid2Mouseup}
  if (button=mbLeft)and (not (ssShift in SHift)) then StringGrid2Click(sender)
  else
  If  (button = mbRight) or (ssshift in shift) then
  with stringgrid2 do
  begin
    if not (mode=usersolving) then resetsolution;
    mode:=usersolving;
    mousetocell(x,y,acol,arow);
    s:=cells[acol,arow];
    if length(s)=1  then
    begin
      s:=s+'*';
      //inc(userstarcount);
    end
    else if (length(s)=2) and (s[2]='*') then
    begin
      s:=s[1];
      //dec(userstarcount);
    end;
    cells[acol,arow]:=s;
    If Issolved then ;//mode:=unknown;
  end;
end;

{**************** LoadCaseList *************}
procedure Tflist.LoadCaselist;
{load the list of avaialble cases initiallially or after adding a case}
var
  ini:Tinifile;
begin
  ini:=TInifile.create(dir+'StarsInGrid.ini');
  with ini do
  begin
    readsections(caseinbox.Items);
    ini.Free;
  end;
  caseinbox.ItemIndex:=0;
end;

{************* CaseInBoxClick **********}
procedure Tflist.CaseInBoxClick(Sender: TObject);
begin  {load the selected case}
  casename:= caseinBox.text;
  loadcase(casename);
end;

end.
